/**
 * @file fbt_dso.c
 * Interface to the shared library loader
 *
 * Copyright (c) 2011 ETH Zurich
 * @author Mathias Payer <mathias.payer@nebelwelt.net>
 *
 * $Date: 2011-03-18 19:09:01 +0100 (Fri, 18 Mar 2011) $
 * $LastChangedDate: 2011-03-18 19:09:01 +0100 (Fri, 18 Mar 2011) $
 * $LastChangedBy: payerm $
 * $Revision: 428 $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */

#define _GNU_SOURCE

#include "fbt_dso.h"

#include <assert.h>
#include <link.h>
#include <elf.h>
#include <sys/types.h>
#include <asm/stat.h>

#include "fbt_datatypes.h"
#include "fbt_libc.h"
#include "fbt_llio.h"
#include "fbt_mem_mgmt.h"
#include "libfastbt.h"

#if defined(VERIFY_CFTX)

/**
 * Function that swaps two symbols
 * @param x first symbol
 * @param y second symbol
 */
static void swap(struct sh_symbol *x, struct sh_symbol *y)
{
  struct sh_symbol temp;
  temp.start = x->start;
  temp.size = x->size;
  temp.type = x->type;
  x->start = y->start;
  x->size = y->size;
  x->type = y->type;
  y->start = temp.start;
  y->size = temp.size;
  y->type = temp.type;
}

/**
 * Sorts all elements of the given sy_symbol array.
 * This is a standard quicksort implementation. Pivot is the (m+n)/2 element.
 * @param list pointer to the symbol array
 * @param m left element
 * @param n right element
 */
static void quicksort(struct sh_symbol *list, ulong_t m, ulong_t n)
{
  ulong_t key,i,j,k;
  if( m < n) {
    k = (m + n) / 2;
    swap(&list[m], &list[k]);
    key = (ulong_t)(list[m].start);
    i = m+1;
    j = n;
    while(i <= j) {
      while((i <= n) && ((ulong_t)(list[i].start) <= key)) i++;
      while((j >= m) && ((ulong_t)(list[j].start) > key)) j--;
      if( i < j) swap(&list[i], &list[j]);
    }
    swap(&list[m], &list[j]);
    // recursively sort the lesser list
    quicksort(list, m, j-1);
    quicksort(list, j+1 ,n);
  }
}

/**
 * Callback used during the scan of the DSOs
 * This callback handles a single DSO. The callback maps the shared library into
 * memory, reads all necessary data (elf header, program table, and section
 * table). The information of the plt, got, got.plt is stored in the dso_chain
 * element for the current DSO. Additionally the dynsym and symbol table are
 * scanned (if available) and all symbols are added to the symbol array for the
 * current DSO.
 * See the man page of dl_iterate_phdr for more information about the
 * parameters.
 * @param info pointer to the struct owned by the loader that contains the
 * information about the current DSO
 * @param size size of the struct (unused here)
 * @param data our data pointer (to the tld)
 */
static int callback(struct dl_phdr_info *info,
                    size_t size __attribute__((unused)), void *data)
{
  struct thread_local_data *tld = (struct thread_local_data*)data;
  
  /* check if this library has already been processed */
  struct dso_chain *cur = tld->dso_objects;

  while (cur != NULL) {
    /* do we already have this dso in our list? */
    if (cur->name == info->dlpi_name) return 0;
    cur = cur->next;  
  }

  struct dso_chain *new = (struct dso_chain*)
    fbt_smalloc(tld, sizeof(struct dso_chain));

  /* store the new DSO at the end of the list */
  if (tld->dso_objects == NULL) {
    tld->dso_objects = new;
  } else {
    cur = tld->dso_objects;
    while (cur->next != NULL) cur = cur->next;
    cur->next = new;
  }
  
  new->name = info->dlpi_name;
  new->baseaddr = (char*) info->dlpi_addr;
  new->endaddr = (char*) (info->dlpi_addr +
                          info->dlpi_phdr[info->dlpi_phnum - 1].p_vaddr + 
                          info->dlpi_phdr[info->dlpi_phnum - 1].p_memsz);
  new->next = NULL;

  //llprintf("name=%s (%d segments)\n", info->dlpi_name, info->dlpi_phnum);
#if defined(PRINT_CFTX)
  if (fbt_strncmp(new->name, "", 1) == 0) {
    llprintf("subgraph \"%p\" { \nlabel=\"/proc/self/exe\"\n", new->baseaddr);
  } else {
    llprintf("subgraph \"%p\" { \nlabel=\"%s\"\n", new->baseaddr, new->name);
  }
#endif
  
  if (fbt_strncmp(new->name, LIBNAME, LIBLEN) != 0) {
    /* get information about plt, got, and individual symbols */ 
    /* open and mmap file */
    int fd, res;
    if (fbt_strncmp(new->name, "", 1) == 0) {
      fbt_open("/proc/self/exe", O_RDONLY, 0, fd);
    } else {
      fbt_open(new->name, O_RDONLY, 0, fd);
    }
    assert((fd & 0xffffff00) != 0xffffff00);

    struct stat fsstat;
    /* the fsstat from the syscall is not the same as in the
       glibc function, so we only allocate some space for the result */
    fbt_fstat(fd, &fsstat, res);

    void *addr;
    fbt_mmap(NULL, fsstat.st_size, PROT_READ, MAP_PRIVATE, fd, 0, addr);
    assert((((int)addr) & 0xffffff00) != 0xffffff00);
    
    /* go to section header table */
    Elf32_Ehdr *header = (Elf32_Ehdr*)addr;
    if (header->e_shnum != 0) {
      Elf32_Shdr *section_header = (Elf32_Shdr*) (addr + header->e_shoff);
      /* iterate through all sections */
      const char *dynstrtab = NULL;
      const char *strtab = NULL;
      Elf32_Shdr *shtdynsym = NULL;
      Elf32_Shdr *shtsymtab = NULL;
      
      int i;
      for (i = 0; i < header->e_shnum; ++i) {

        if (header->e_shstrndx != SHN_UNDEF) {
          const char *name_entry = (const char*)
            (addr + section_header[header->e_shstrndx].sh_offset
             + section_header[i].sh_name);
          /* TODO: what about .rel.plt ? */
          if (fbt_strncmp(name_entry, ".plt", 5) == 0) {
            new->plt = (char*)(info->dlpi_addr + section_header[i].sh_addr);
            new->plt_size = section_header[i].sh_size;
          }
          if (fbt_strncmp(name_entry, ".got.plt", 9) == 0) {
            new->gotplt = (char*)(info->dlpi_addr + section_header[i].sh_addr);
            new->gotplt_size = section_header[i].sh_size;
          }
          if (fbt_strncmp(name_entry, ".got", 5) == 0) {
            new->got = (char*)(info->dlpi_addr + section_header[i].sh_addr);
            new->got_size = section_header[i].sh_size;
          }
          if (section_header[i].sh_type == SHT_STRTAB &&
              fbt_strncmp(name_entry, ".dynstr", 8) == 0) {
            /* dynamic string table */
            dynstrtab = addr + section_header[i].sh_offset;
          }
          if (section_header[i].sh_type == SHT_STRTAB &&
              fbt_strncmp(name_entry, ".strtab", 8) == 0) {
            /* symbol string table */
            strtab = addr + section_header[i].sh_offset;
          }
        }  /* valid section */
        if (section_header[i].sh_type == SHT_SYMTAB) {  /* symbol table */
          shtsymtab = &section_header[i];
        }
        if (section_header[i].sh_type == SHT_DYNSYM) {  /* dynamic symbol table */
          shtdynsym = &section_header[i];
        }
      }  /* loop through sections */
      
      if (shtdynsym != NULL && dynstrtab != NULL) {        
        Elf32_Sym *dynsym = (Elf32_Sym*)(info->dlpi_addr + shtdynsym->sh_addr);
        long nsyms = shtdynsym->sh_size / sizeof(Elf32_Sym);

        long locsyms = 0;
        int j, k;
        for (j = 0; j < nsyms; ++j) {
          if ((ELF32_ST_TYPE(dynsym[j].st_info) == STT_FUNC) &&
              (dynsym[j].st_shndx != SHT_NULL)) ++locsyms; 
          if (ELF32_ST_TYPE(dynsym[j].st_info) == STT_FUNC) ++locsyms;
        }

        if (shtsymtab != NULL) {
          Elf32_Sym *symtab = (Elf32_Sym*)(addr + shtsymtab->sh_offset);
          long nssyms = shtsymtab->sh_size / sizeof(Elf32_Sym);
          for (j = 0; j < nssyms; ++j) {
            if ((ELF32_ST_TYPE(symtab[j].st_info) == STT_FUNC) &&
                (symtab[j].st_shndx != SHT_NULL)) ++locsyms; 
            if (ELF32_ST_TYPE(symtab[j].st_info) == STT_FUNC) ++locsyms;
          }
        }
        
        struct sh_symbol *syms = (struct sh_symbol*)
          fbt_lalloc(tld, NRPAGES(locsyms * sizeof(struct sh_symbol)),
                     MT_INTERNAL );
        new->nr_symbols = locsyms;
        new->symbols = syms;
        /* we got a pointer to the symbols, so iterate through them and add
           them to our list */
        for (j = 0, k = 0; j < nsyms; ++j) {
          assert(k < locsyms);
          const char *symname = (const char*)(dynstrtab + dynsym[j].st_name);
          if ((ELF32_ST_TYPE(dynsym[j].st_info) == STT_FUNC) &&
              (dynsym[j].st_shndx != SHT_NULL)) {            
            /* defined function */
            syms[k].size = dynsym[j].st_size;
            syms[k].start = (char*)info->dlpi_addr + dynsym[j].st_value;
            syms[k].type = STT_LOCAL;
            llprintf("\"%p\" [label=\"%s\"];\n", syms[k].start, symname);
            ++k;
          } else
          if (ELF32_ST_TYPE(dynsym[j].st_info) == STT_FUNC) {
            /* external function in other dso */
            syms[k].size = dynsym[j].st_size;
            syms[k].start = NULL;
            syms[k].type = STT_REF;
            //llprintf("\"%pxx\" [label=\"%s\"];\n", new->baseaddr, symname);
            ++k;
          }
        }  /* parse dynsym information */

        /* get symtab information */
        if (shtsymtab != NULL) {
          Elf32_Sym *symtab = (Elf32_Sym*)(addr + shtsymtab->sh_offset);
          long nssyms = shtsymtab->sh_size / sizeof(Elf32_Sym);
          for (j = 0; j < nssyms; ++j) {
            assert(k < locsyms);
#if defined(PRINT_CFTX)
            const char *symname = (const char*)(strtab + symtab[j].st_name);
#endif
            if ((ELF32_ST_TYPE(symtab[j].st_info) == STT_FUNC) &&
                (symtab[j].st_shndx != SHT_NULL)) {            
              /* defined function */
              syms[k].size = symtab[j].st_size;
              /* static objects store absolute value */
              syms[k].start = (char*)symtab[j].st_value; 
              syms[k].type = STT_LOCAL;
#if defined(PRINT_CFTX)
              llprintf("\"%p\" [label=\"%s\"];\n", syms[k].start, symname);
#endif
              ++k;
            } else
              if (ELF32_ST_TYPE(symtab[j].st_info) == STT_FUNC) {
                /* external function in other dso */
                syms[k].size = symtab[j].st_size;
                syms[k].start = NULL;
                syms[k].type = STT_REF;
#if defined(PRINT_CFTX)
                //llprintf("\"%pxx\" [label=\"%s\"];\n", new->baseaddr, symname);
#endif
                ++k;
              }
          }
        }  /* parse symtabl information */
        /* now we must sort the symbol table of this DSO */
        quicksort(syms, 0, locsyms);
        for (i = 0; i < locsyms-1; ++i) {
          if (syms[i].type == STT_LOCAL && syms[i].size == 0 &&
              syms[i].start != NULL) {
            syms[i].size = syms[i+1].start - syms[i].start - 1;
          }
        }
      } /* extract information of DSO */
      
      /* TODO: fail if no information available */
    }  /* valid section table */
    
#if defined(PRINT_CFTX)
    llprintf("}\n");
#endif
    
    /* close library again */
    fbt_munmap(addr, fsstat.st_size, res);
    fbt_close(fd, res);
    /* be merry */

    /* TODO: don't forget to catch all dlopen dlclose calls */
  };  /* if new->name != libfastbt.so */
  
  return 0;
}

void fbt_check_transfer(struct thread_local_data *tld, unsigned char *src,
                        unsigned char *dst, enum cftx_type type)
{
  if (type == CFTX_JMP_IND || type == CFTX_CALL_IND || type == CFTX_RET) {
    /* dst is only a ptr to a ptr, so we must dereference once */
    dst = (unsigned char*)(*((long*)dst));
  }
  struct dso_chain *cur = tld->dso_objects;
  struct dso_chain *srcdso = NULL, *dstdso = NULL;
  long src_sym = -1, dst_sym = -1;
  struct sh_symbol *src_symbol = NULL, *dst_symbol = NULL;
  while (cur != NULL) {
    if (PTR_IN_REGION(src, cur->baseaddr,
                      (long)(cur->endaddr - cur->baseaddr))) {
      srcdso = cur;
      int i;
      for (i = 0; i < cur->nr_symbols; ++i) {
        struct sh_symbol *cur_symbol = &(cur->symbols[i]);
        if (PTR_IN_REGION(src, cur_symbol->start, cur_symbol->size)) {
          src_sym = i;
          src_symbol = cur_symbol;
          break;
        }
      }  /* for each symbol in dso */
      if (PTR_IN_REGION(src, cur->plt, cur->plt_size)) {
        src_sym = (((unsigned long)src - (unsigned long)cur->plt)/16)*(-1) - 2;
      }
    }
    if (PTR_IN_REGION(dst, cur->baseaddr,
                      (long)(cur->endaddr - cur->baseaddr))) {
      dstdso = cur;
      int i;
      for (i = 0; i < cur->nr_symbols; ++i) {
        struct sh_symbol *cur_symbol = &(cur->symbols[i]);
        if (PTR_IN_REGION(dst, cur_symbol->start, cur_symbol->size)) {
          dst_sym = i;
          dst_symbol = cur_symbol;
          break;
        }
      }  /* for each symbol in dso */
      if (PTR_IN_REGION(dst, cur->plt, cur->plt_size)) {
        dst_sym = (((unsigned long)dst - (unsigned long)cur->plt)/16)*(-1) - 2;
      }
    }
    cur = cur->next;
  }  /* iterate all dsos */
  
#if defined(PRINT_CFTX)
  if (srcdso != NULL && dstdso != NULL && dst_sym != -1) {
    if (src_sym <= -1) {
      llprintf("\"%p.%d\" -> ", srcdso->baseaddr, src_sym);
    } else {
      llprintf("\"%p\" -> ", src_symbol->start);
    }
    if (dst_sym <= -2) {
      llprintf("\"%p.%d\"\n", dstdso->baseaddr, dst_sym);
    } else {
      llprintf("\"%p\"\n", dst_symbol->start);
    }
  }
#endif
  /* TODO: fail if the control flow transfer is not valid */
}

void fbt_rescan_dsos(struct thread_local_data *tld)
{
  dl_iterate_phdr(callback, tld);
}

#endif  /* VERIFY_CFTX */
